const xml2js = require('xml2js');
const parseSignature = require('../signature');
const ProxyInterface = require('./proxy-interface');
const variant = require('../service/variant');
const Variant = variant.Variant;
const { DBusError } = require('../errors');
const {
  assertInterfaceNameValid,
  assertObjectPathValid,
  isObjectPathValid,
  isInterfaceNameValid
} = require('../validators');

/**
 * @class
 * A class that represents a proxy to a DBus object. The `ProxyObject` contains
 * `ProxyInterface`s and a list of `node`s which are object paths of child
 * objects. A `ProxyObject` is created through {@link
 * MessageBus#getProxyObject} for a given well-known name and object path.
 * An interface can be gotten through {@link ProxyObject#getInterface} and can
 * be used to call methods and receive signals for that interface.
 *
 * @example
 * let bus = dbus.sessionBus();
 * let obj = await bus.getProxyObject('org.freedesktop.DBus', '/org/freedesktop/DBus');
 * let peer = obj.getInterface('org.freedesktop.DBus.Peer')
 * await peer.Ping();
 */
class ProxyObject {
  /**
   * Create a new `ProxyObject`. This constructor should not be called
   * directly. Use {@link MessageBus#getProxyObject} to get a proxy object.
   */
  constructor(bus, name, path) {
    assertInterfaceNameValid(name);
    assertObjectPathValid(path);
    /**
     * The {@link MessageBus} this name belongs to.
     * @memberof ProxyObject#
     * @member {MessageBus} bus
     */
    this.bus = bus;
    /**
     * The well-known bus name for this proxy object as a string.
     * @memberof ProxyObject#
     * @member {string} name
     */
    this.name = name;
    /**
     * The object path for this `ProxyObject`.
     * @memberof ProxyObject#
     * @member {string} path
     */
    this.path = path;
    /**
     * The object path child nodes for this `ProxyObject` as an array of
     * strings
     * @memberof ProxyObject#
     * @member {string[]} nodes
     */
    this.nodes = [];

    /**
     * A map of interface names to [ProxyInterfaces]{@link ProxyInterface} for
     * this `ProxyObject`.
     * @memberof ProxyObject#
     * @member {Object.<string, ProxyInterface>} interfaces
     */
    this.interfaces = {};
    this._parser = new xml2js.Parser();
  }

  /**
   * Get a {@link ProxyInterface} for the given interface name.
   *
   * @param name {string} - the interface name to get.
   * @returns {ProxyInterface} - the proxy interface with this name exported by
   * the object or `undefined` if the object does not export an interface with
   * that name.
   * @throws {Error} Throws an error if the interface is not found on this object.
   */
  getInterface(name) {
    if (!Object.keys(this.interfaces).includes(name)) {
      throw new Error(`interface not found in proxy object: ${name}`);
    }
    return this.interfaces[name];
  }

  _initXml(xml) {
    let root = xml.node;

    if (Array.isArray(root.node)) {
      for (let n of root.node) {
        if (!n.hasOwnProperty('$')) {
          continue;
        }
        let name = n['$'].name;
        let path = `${this.path}/${name}`;
        if (isObjectPathValid(path)) {
          this.nodes.push(path);
        }
      }
    }

    if (Array.isArray(root.interface)) {
      for (let i of root.interface) {
        let iface = ProxyInterface._fromXml(this, i);
        if (iface !== null) {
          this.interfaces[iface.$name] = iface;
          this.bus._addMatch(`type='signal',sender=${this.name},interface='${iface.$name}',path='${this.path}'`);
          for (let signal of iface.$signals) {
            let event = JSON.stringify({
              path: this.path,
              interface: iface.$name,
              member: signal.name
            });
            this.bus._signals.on(event, (msg) => {
              let {body, signature, sender} = msg;
              if (this.bus._nameOwners[this.name] !== sender) {
                return;
              }
              if (signature !== signal.signature) {
                console.error(`warning: got signature ${signature} for signal ${iface.$name}.${signal.name} (expected ${signal.signature})`);
                return;
              }
              // TODO refactor this into a method
              let result = [];
              let signatureTree = parseSignature(signature);
              for (let i = 0; i < signatureTree.length; ++i) {
                let tree = signatureTree[i];
                result.push(variant.parse([[tree], [body[i]]]));
              }
              iface.emit.apply(iface, [signal.name].concat(result));
            });
          }
        }
      }
    }
  }

  _init() {
    let introspect = new Promise((resolve, reject) => {
      this.bus._invoke(
        {
          destination: this.name,
          path: this.path,
          interface: 'org.freedesktop.DBus.Introspectable',
          member: 'Introspect',
          signature: '',
          body: []
        },
        (err, xml) => {
          if (err) {
            reject(err);
            return;
          }

          this._parser.parseString(xml, (err, data) => {
            if (err) {
              reject(err);
              return;
            }
            this._initXml(data);
            resolve(this);
          });
        }
      );
    });

    return new Promise((resolve, reject) => {
      Promise.all([introspect, this.bus._getNameOwner(this.name)])
        .then((result) => {
          return resolve(result[0]);
        })
        .catch((err) => {
          return reject(err);
        });
    });
  }

  _callMethod(iface, member, inSignature, outSignature, ...args) {
    return new Promise((resolve, reject) => {
      args = args || [];

      // TODO refactor this into a method
      let inSignatureTree = parseSignature(inSignature);
      let body = [];
      for (let i = 0; i < args.length; ++i) {
        if (inSignatureTree[i].type === 'v') {
          if (args[i].constructor !== Variant) {
            throw new Error(`method call ${iface}.${member} expected a Variant() argument for arg ${i+1} (value='${args[i]}')`);
          }
          body.push(variant.jsToMarshalFmt(args[i].signature, args[i].value));
        } else {
          body.push(variant.jsToMarshalFmt(inSignatureTree[i], args[i])[1]);
        }
      }

      let msg = {
        member: member,
        signature: inSignature,
        destination: this.name,
        path: this.path,
        interface: iface,
        body: body
      };

      this.bus._invoke(msg, function(err, ...busResult) {
        if (err) {
          if (this.message && this.message.hasOwnProperty('errorName')) {
            reject(new DBusError(this.message.errorName, err[0]));
          } else {
            reject(err);
          }
          return;
        }
        // TODO refactor this into a method
        let result = [];
        let outSignatureTree = parseSignature(outSignature);
        if (outSignatureTree.length === 0) {
          resolve(null);
          return;
        }
        for (let i = 0; i < outSignatureTree.length; ++i) {
          let tree = outSignatureTree[i];
          result.push(variant.parse([[tree], [busResult[i]]]));
        }
        if (outSignatureTree.length === 1) {
          resolve(result[0]);
        } else {
          resolve(result);
        }
      });
    });
  }
}

module.exports = ProxyObject;
